VERSION 5.00
Begin VB.UserControl Mailslot 
   Appearance      =   0  'Flat
   BackColor       =   &H80000005&
   BorderStyle     =   1  'Fixed Single
   CanGetFocus     =   0   'False
   ClientHeight    =   480
   ClientLeft      =   0
   ClientTop       =   0
   ClientWidth     =   480
   ClipBehavior    =   0  'None
   ClipControls    =   0   'False
   HasDC           =   0   'False
   HitBehavior     =   0  'None
   InvisibleAtRuntime=   -1  'True
   Picture         =   "Mailslot.ctx":0000
   ScaleHeight     =   479.135
   ScaleMode       =   0  'User
   ScaleWidth      =   479.135
   ToolboxBitmap   =   "Mailslot.ctx":03FB
   Begin VB.Timer tmrPoll 
      Enabled         =   0   'False
      Interval        =   1000
      Left            =   0
      Top             =   0
   End
End
Attribute VB_Name = "Mailslot"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'
'========
'Mailslot
'========
'
'OVERVIEW
'========
'
'This Mailslot control is meant to be used by peer machines
'to communicate within a local area network.
'
'Unlike TCP/IP socket based solutions, Mailslot does not
'require a dedicated TCP or UDP port.  It only requires
'that normal NetBIOS traffic over TCP/IP, IPX, or NetBEUI
'is available among the systems in the Domain or Workgroup.
'
'No dedicated server is needed, and no server components or
'services need to be provisioned.
'
'
'RESTRICTIONS
'============
'
'The Mailslot control uses NetBIOS Mailslots for communication
'within the Domain or Workgroup.  Thus it isn't suitable for
'use over the wider Internet.
'
'All communicating computers must be in the same Domain or
'Workgroup.
'
'Due to the way NetBIOS Mailslots work only one running
'instance of any application on a given machine can monitor a
'given combination of AppID & ResourceID.  I.e. an instance of
'Mailslot that is listening on a given name is a Mailslot
'Server for that name, and these must be unique per machine.
'THIS MAY MAKE THE MAILSLOT CONTROL IMPRACTICAL FOR
'TERMINAL SERVER OR CITRIX APPLICATIONS!
'
'Systems running software firewalls need to allow NetBIOS
'traffic through.  In most cases you can grant an exception
'for this traffic by allowing "file and print sharing" in
'your software firewall.
'
'Names are limited to 8 character nodes in order to make sure
'Mailslot works in Win9x machines.  We use Timer-based polling
'instead of asynchronous callbacks for the same reason.
'
'NORMAL USAGE PATTERN
'====================
'
'Drop one or more instances of Mailslot on a form, possibly as
'a control array.  Use only one instance of Mailslot for each
'named Mailslot to be monitored.
'
'Set properties as desired:
'
'   AutoPad
'   PollInterval
'
'Call Initialize() first if you want to access ComputerName
'or UserName before calling Listen() the first time.
'
'Begin listening for messages by invoking the Listen()
'method with an appropriate Name value.
'
'Respond to MsgArrival() events by calling GetMsg().
'
'Send messages by invoking the SendMsg() method.
'
'Stop listening by invoking the CloseSlot() method when
'complete.
'
'After a CloseSlot() the program can invoke Listen() again
'with the same or a new name if desired.
'
'DEPENDENCIES
'============
'
'VB6 Runtime.  Only system APIs and the VB Runtime instrinsic
'Timer control are used here.
'
'CLASS MEMBERS
'=============
'
'   PROPERTIES
'   ----------
'
'   AutoPad         (Boolean, R/W)
'
'                   If set True then SendMsg() calls with Msg
'                   values of 425 or 426 are padded with NULs
'                   to 427 characters.  Otherwise these Msg
'                   values cause an exception.
'
'                   Not applicable to broadcast (Computer = "*")
'                   requests, which by nature are limited to 424
'                   characters.
'
'                   DEFAULT: False.
'
'   ComputerName    (String, RO)
'
'                   Returns the NetBIOS name of the local
'                   system.  For any general use the program
'                   may make of it.
'
'                   Only valid once Initialize() has been
'                   called.
'
'   PollInterval    (Long, R/W)
'
'                   How long (in millseconds) to check for new
'                   messages.  Valid range is 100 to 30000.
'
'                   DEFAULT: 500 milliseconds.
'
'   UserName        (String, RO)
'
'                   Returns the logon user name of the Windows
'                   session.  For any general use the program
'                   may make of it.
'
'                   Only valid once Initialize() has been
'                   called.
'
'   METHODS
'   -------
'
'   CloseSlot()     Stops listening for messages.  Should be
'                   called as part of unloading the parent
'                   form.
'
'   GetMsg(ByRef Msg As String) As Boolean
'
'                   Reads the current message into Msg.  Returns
'                   True upon success, False if called when no
'                   message is present.
'
'                   Should normally be called from MsgArrival
'                   event handler and called repeatedly until
'                   False.
'
'   Initialize()    Must be called once, automatically called
'                   by first Listen() call.
'
'   Listen(ByVal Name As String)
'
'                   Starts listening for messages.  Name must
'                   be made up of nodes of 1 to 8 characters,
'                   alphanumeric, separated by "\" characters.
'                   These values are case insensitive.
'
'                   Mailslot must be Listen()ing in order to
'                   call GetMsg() and SendMsg().
'
'   SendMsg(ByVal Computer As String, ByVal Msg As String)
'
'                   Sends Msg to Computer.  If Computer is "*"
'                   then Msg is broadcast to the whole network.
'
'                   NOTE:
'
'                   Broadcast sends are limited to 424
'                   characters.  No message may be exactly 425
'                   or 426 characters, so pad longer messages
'                   or set AutoPad = True.
'
'                   Violating either constraint will result in
'                   an exception.
'
'   EVENTS
'   ------
'
'   MsgArrival()    Occurs whenever a message is received.
'
'EXCEPTIONS SOURCE       DESCRIPTION
'========== ============ =======================================
'
'&H8004C700 PollTimer    GetMailslotInfo failed (System Err #)
'&H8004C710 CloseSlot    CloseHandle failed (System Err #)
'&H8004C720 Listen       Invalid Name parameter value
'&H8004C740 Listen       CreateMailslot failed, mailslot in use
'&H8004C750 Listen       CreateMailslot failed (System Err #)
'&H8004C760 SendMsg      CreateFile failed (System Err #)
'&H8004C770 SendMsg      WriteFile failed (System Err #)
'&H8004C780 SendMsg      CloseHandle failed (System Err #)
'&H8004C790 PollInterval PollInterval set outside valid range
'&H8004C7A0 GetMsg       ReadFile failed (System Err #)
'

Private Const MYPREFIX As String = "\\.\mailslot\"
Private Const BROADCASTPREFIX As String = "\\*\mailslot\"
Private Const NAMEPATTERN = "[!0-9A-Za-z]"
Private Const DEFAULTPOLLINTERVAL As Long = 500
Private Const SHORTNAMELENGTH As Long = 8
Private Const MAX_COMPUTERNAME_LENGTH = 31
Private Const MAX_USERNAME_LENGTH = 254

Private Const ERROR_ALREADY_EXISTS = 183
Private Const FILE_ATTRIBUTE_NORMAL = &H80
Private Const FILE_SHARE_READ = &H1
Private Const GENERIC_WRITE = &H40000000
Private Const INVALID_HANDLE_VALUE = (-1)
Private Const MAILSLOT_NO_MESSAGE = (-1)
Private Const MAILSLOT_WAIT_FOREVER = (-1)
Private Const OPEN_EXISTING = 3

Private Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long) As Long

Private Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" ( _
    ByVal lpFileName As String, _
    ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, _
    lpSecurityAttributes As Any, _
    ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long

Private Declare Function CreateMailslot Lib "kernel32" Alias "CreateMailslotA" ( _
    ByVal lpName As String, _
    ByVal nMaxMessageSize As Long, _
    ByVal lReadTimeout As Long, _
    lpSecurityAttributes As Any) As Long

Private Declare Function GetComputerName Lib "kernel32" Alias "GetComputerNameA" ( _
    ByVal lpBuffer As String, _
    nSize As Long) As Long

Private Declare Function GetMailslotInfo Lib "kernel32" ( _
    ByVal hMailslot As Long, _
    lpMaxMessageSize As Long, _
    lpNextSize As Long, _
    lpMessageCount As Long, _
    lpReadTimeout As Long) As Long

Private Declare Function GetUserName Lib "advapi32" Alias "GetUserNameA" ( _
    ByVal lpBuffer As String, _
    nSize As Long) As Long

Private Declare Function ReadFile Lib "kernel32" ( _
    ByVal hFile As Long, _
    lpBuffer As Any, _
    ByVal nNumberOfBytesToRead As Long, _
    lpNumberOfBytesRead As Long, _
    lpOverlapped As Any) As Long

Private Declare Function WriteFile Lib "kernel32" ( _
    ByVal hFile As Long, _
    lpBuffer As Any, _
    ByVal nNumberOfBytesToWrite As Long, _
    lpNumberOfBytesWritten As Long, _
    lpOverlapped As Any) As Long

Private hServer As Long
Private strBroadcastMailslot As String
Private strComputerName As String
Private strUserName As String
Private colInboundQ As New Collection
Private lngPollInterval As Long
Private blnAutoPad As Boolean
Private blnInitialized As Boolean

Public Event MsgArrival()

Public Property Let AutoPad(ByVal YesNo As Boolean)
    blnAutoPad = YesNo
End Property

Public Property Get AutoPad() As Boolean
    AutoPad = blnAutoPad
End Property

Public Property Get ComputerName() As String
Attribute ComputerName.VB_Description = "NetBIOS name of this computer."
    ComputerName = strComputerName
End Property

Public Property Let PollInterval(ByVal MilliSeconds As Long)
    If 100 <= MilliSeconds And MilliSeconds <= 30000 Then
        lngPollInterval = MilliSeconds
        tmrPoll.Interval = MilliSeconds
    Else
        Err.Raise &H8004C790, "Mailslot.PollInterval", _
                  "PollInterval must be 100 to 30000 milliseconds"
    End If
End Property

Public Property Get PollInterval() As Long
    PollInterval = lngPollInterval
End Property

Public Property Get UserName() As String
    UserName = strUserName
End Property

Public Sub CloseSlot()
Attribute CloseSlot.VB_Description = "Stops listening for alerts, prepares for new Listen() calls."
    If blnInitialized Then
        If hServer <> INVALID_HANDLE_VALUE Then
            If CloseHandle(hServer) = 0 Then
                Err.Raise &H8004C710, "Mailslot.CloseSlot", _
                          "CloseHandle failed, error " & CStr(Err.LastDllError)
                hServer = INVALID_HANDLE_VALUE
            Else
                tmrPoll.Enabled = False
            End If
        End If
    End If
End Sub

Public Function GetMsg(ByRef Msg As String) As Boolean
    'If there are messages read the next one and return True,
    'otherwise return False.
    
    With colInboundQ
        If .Count > 0 Then
            Msg = StrConv(.Item(1), vbUnicode)
            .Remove 1
            GetMsg = True
        End If
    End With
End Function

Private Sub GetMSInfo(ByRef cbMessage As Long, ByRef cMessage As Long)
    If GetMailslotInfo(hServer, ByVal 0&, cbMessage, cMessage, ByVal 0&) = 0 Then
        Err.Raise &H8004C700, "Mailslot.PollTimer", _
                  "GetMailslotInfo failed, error " & CStr(Err.LastDllError)
    End If
End Sub

Public Sub Initialize()
    Dim nSize As Long
    
    hServer = INVALID_HANDLE_VALUE
    
    nSize = MAX_COMPUTERNAME_LENGTH + 1
    strComputerName = String$(nSize, 0)
    If GetComputerName(strComputerName, nSize) = 0 Then
        strComputerName = ""
    Else
        strComputerName = Left$(strComputerName, nSize)
    End If
    
    nSize = MAX_USERNAME_LENGTH + 1
    strUserName = String$(nSize, 0)
    If GetUserName(strUserName, nSize) = 0 Then
        strUserName = ""
    Else
        strUserName = Left$(strUserName, nSize - 1)
    End If
    
    blnInitialized = True
End Sub

Private Function InvalidName(ByVal Name As String) As Boolean
    Dim strNodes() As String
    Dim I As Long
    Dim J As Long
    
    strNodes = Split(Name, "\")
    For I = 0 To UBound(strNodes)
        If 1 <= Len(strNodes(I)) And Len(strNodes(I)) <= SHORTNAMELENGTH Then
            For J = 1 To Len(strNodes(I))
                If Mid$(strNodes(I), J, 1) Like NAMEPATTERN Then
                    InvalidName = True
                    Exit Function
                End If
            Next
        Else
            InvalidName = True
            Exit Function
        End If
    Next
End Function

Public Sub Listen(ByVal Name As String)
Attribute Listen.VB_Description = "Begins listening for alerts sent to AppID, ResourceID."
    'Validate Name.  Construct mailslot names for server
    'and for broadcast writes.  Start mailslot server and
    'poll interval timer.
    Dim strServerMailslot As String
    Dim lngDLLError As Long

    If InvalidName(Name) Then
        Err.Raise &H8004C720, "Mailslot.Listen", _
                  "Invalid Name parameter value"
    End If
    If Not blnInitialized Then Initialize
    strServerMailslot = MYPREFIX & Name
    strBroadcastMailslot = BROADCASTPREFIX & Name
    
    If hServer <> INVALID_HANDLE_VALUE Then
        CloseSlot
    End If
    
    hServer = CreateMailslot(strServerMailslot, 0&, _
                             MAILSLOT_WAIT_FOREVER, ByVal 0&)
    If hServer = INVALID_HANDLE_VALUE Then
        lngDLLError = Err.LastDllError
        If lngDLLError = ERROR_ALREADY_EXISTS Then
            Err.Raise &H8004C740, "Mailslot.Listen", _
                      "CreateMailslot failed, mailslot in use"
        Else
            Err.Raise &H8004C750, "Mailslot.Listen", _
                      "CreateMailslot failed, error " & CStr(lngDLLError)
        End If
    Else
        tmrPoll.Enabled = True
    End If
End Sub

Public Sub SendMsg(ByVal Computer As String, ByVal Msg As String)
Attribute SendMsg.VB_Description = "Sends an alert."
    Dim strMailslot As String
    Dim hClient As Long
    Dim bytMsg() As Byte
    Dim cbWritten As Long
    
    Computer = Trim$(Computer)
    If Computer = "*" Then
        If Len(Msg) > 424 Then
            Err.Raise &H8004C752, "Mailslot.SendMsg", _
                      "Can't broadcast messages longer than 424 chars"
        End If
        strMailslot = strBroadcastMailslot
    Else
        If Len(Msg) = 425 Or Len(Msg) = 426 Then
            If blnAutoPad Then
                If Len(Msg) = 425 Then Msg = Msg & vbNullChar
                Msg = Msg & vbNullChar
            Else
                Err.Raise &H8004C754, "Mailslot.SendMsg", _
                          "Can't send messages of 425 or 426 chars"
            End If
        End If
        strMailslot = Replace$(strBroadcastMailslot, "*", Computer)
    End If
    
    hClient = CreateFile(strMailslot, GENERIC_WRITE, _
                         FILE_SHARE_READ, ByVal 0&, OPEN_EXISTING, _
                         FILE_ATTRIBUTE_NORMAL, ByVal 0&)
    If hClient = INVALID_HANDLE_VALUE Then
        Err.Raise &H8004C760, "Mailslot.SendMsg", _
                  "CreateFile failed, error " & CStr(Err.LastDllError)
    Else
        bytMsg = StrConv(Msg, vbFromUnicode)
        If WriteFile(hClient, bytMsg(0), UBound(bytMsg) + 1, cbWritten, ByVal 0&) = 0 Then
            Err.Raise &H8004C770, "Mailslot.SendMsg", _
                      "WriteFile failed, error " & CStr(Err.LastDllError)
        End If
        If CloseHandle(hClient) = 0 Then
            Err.Raise &H8004C780, "Mailslot.SendMsg", _
                      "CloseHandle failed, error " & CStr(Err.LastDllError)
        End If
    End If
End Sub

Private Sub tmrPoll_Timer()
    'When timer fires check for messages.
    'Signal MsgArrival() if any are available.
    Dim cbMessage As Long
    Dim cMessage As Long
    Dim cbRead As Long
    Dim bytMsg() As Byte

    GetMSInfo cbMessage, cMessage
    If cbMessage <> MAILSLOT_NO_MESSAGE Then
        Do While cMessage > 0
            ReDim bytMsg(cbMessage - 1)
            If ReadFile(hServer, bytMsg(0), cbMessage, cbRead, ByVal 0&) = 0 Then
                Err.Raise &H8004C7A0, "Mailslot.GetMsg", _
                          "ReadFile failed, error " & CStr(Err.LastDllError)
            Else
                colInboundQ.Add bytMsg
            End If
            
            GetMSInfo cbMessage, cMessage
        Loop
        
        RaiseEvent MsgArrival
    End If
End Sub

Private Sub UserControl_InitProperties()
    AutoPad = False
    PollInterval = DEFAULTPOLLINTERVAL
End Sub

Private Sub UserControl_Paint()
    Width = 480
    Height = 480
End Sub

Private Sub UserControl_ReadProperties(PropBag As PropertyBag)
    AutoPad = PropBag.ReadProperty("AutoPad", False)
    PollInterval = PropBag.ReadProperty("PollInterval", DEFAULTPOLLINTERVAL)
End Sub

Private Sub UserControl_Terminate()
    If hServer <> INVALID_HANDLE_VALUE Then
        CloseHandle hServer
    End If
End Sub

Private Sub UserControl_WriteProperties(PropBag As PropertyBag)
    PropBag.WriteProperty "AutoPad", AutoPad, False
    PropBag.WriteProperty "PollInterval", PollInterval, DEFAULTPOLLINTERVAL
End Sub
