Attribute VB_Name = "MainMod"
Option Explicit

Private Const MAILSLOT_PREFIX As String = "\\*\mailslot\LANegram\"
Private Const DEFAULT_CHANNEL As String = "0"

Private Const FILE_ATTRIBUTE_NORMAL = &H80
Private Const FILE_SHARE_READ = &H1
Private Const GENERIC_WRITE = &H40000000
Private Const INVALID_HANDLE_VALUE = -1
Private Const OPEN_EXISTING = 3

Private Const STD_INPUT_HANDLE = -10&
Private Const STD_OUTPUT_HANDLE = -11&

Private Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long) As Long

Private Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" ( _
    ByVal lpFileName As String, _
    ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, _
    lpSecurityAttributes As Any, _
    ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long

Private Declare Function GetComputerName Lib "kernel32" Alias "GetComputerNameA" ( _
    ByVal lpBuffer As String, _
    nSize As Long) As Long

Private Declare Function GetStdHandle Lib "kernel32" ( _
    ByVal nStdHandle As Long) As Long

Private Declare Function GetUserName Lib "advapi32" Alias "GetUserNameA" ( _
    ByVal lpBuffer As String, _
    nSize As Long) As Long

Private Declare Function ReadFile Lib "kernel32" ( _
    ByVal hFile As Long, _
    lpBuffer As Any, _
    ByVal nNumberOfBytesToRead As Long, _
    lpNumberOfBytesRead As Long, _
    lpOverlapped As Any) As Long

Private Declare Function WriteFile Lib "kernel32" ( _
    ByVal hFile As Long, _
    lpBuffer As Any, _
    ByVal nNumberOfBytesToWrite As Long, _
    lpNumberOfBytesWritten As Long, _
    lpOverlapped As Any) As Long

Private blnInteractive As Boolean
Private hStdIn As Long
Private hStdOut As Long
Private strComputerName As String
Private strUserName As String
Private strParts() As String

Private Sub GetNames()
    Const MAX_COMPUTERNAME_LENGTH = 31
    Const MAX_USERNAME_LENGTH = 254
    Dim nSize As Long
    
    nSize = MAX_COMPUTERNAME_LENGTH + 1
    strComputerName = String$(nSize, 0)
    If GetComputerName(strComputerName, nSize) = 0 Then
        strComputerName = ""
    Else
        strComputerName = Left$(strComputerName, nSize)
    End If
    
    nSize = MAX_USERNAME_LENGTH + 1
    strUserName = String$(nSize, 0)
    If GetUserName(strUserName, nSize) = 0 Then
        strUserName = ""
    Else
        strUserName = Left$(strUserName, nSize - 1)
    End If
End Sub

Private Function DestinationCheck(ByVal Pattern As String) As Boolean
    On Error Resume Next
    DestinationCheck = vbTab Like Pattern 'Test Pattern value for failure.
    DestinationCheck = Err.Number = 0
    If Not DestinationCheck Then
        WriteLine "#Bad NAME pattern string"
    End If
End Function

Private Sub ParseAndExecute(ByVal CommandLine As String)
    Dim intMaxParts As Integer
    
    strParts = Split(CommandLine, " ")
    intMaxParts = UBound(strParts)
    If intMaxParts < 1 Then
        Usage
    ElseIf strParts(0) = "/?" Then
        Usage
    ElseIf LCase$(strParts(0)) = "/c" Then
        If intMaxParts < 3 Then
            Usage
        ElseIf DestinationCheck(strParts(2)) Then
            SendMsg strParts(1), strParts(2), 3
        End If
    Else
        If intMaxParts < 1 Then
            Usage
        ElseIf DestinationCheck(strParts(0)) Then
            SendMsg DEFAULT_CHANNEL, strParts(0), 1
        End If
    End If
End Sub

Private Function ReadLine() As String
    Dim cbRead As Long
    Dim bytText() As Byte

    ReDim bytText(500)
    If ReadFile(hStdIn, bytText(0), 500, cbRead, ByVal 0&) = 0 Then
        WriteLine "#ReadFile failed, error " & CStr(Err.LastDllError)
    ElseIf cbRead >= 3 Then
        ReDim Preserve bytText(cbRead - 3)
        ReadLine = StrConv(bytText, vbUnicode)
    End If
End Function

Private Function ReJoinAt(ByVal StartPos As Integer)
    For StartPos = StartPos To UBound(strParts) - 1
        ReJoinAt = ReJoinAt & strParts(StartPos) & " "
    Next
    ReJoinAt = ReJoinAt & strParts(StartPos)
End Function

Private Sub SendMsg( _
    ByVal Channel As String, _
    ByVal Destination As String, _
    ByVal MsgStart As Integer)
    
    Dim hClient As Long
    Dim bytMsg() As Byte
    Dim cbWritten As Long
    
    hClient = CreateFile(MAILSLOT_PREFIX & Channel, GENERIC_WRITE, _
                         FILE_SHARE_READ, ByVal 0&, OPEN_EXISTING, _
                         FILE_ATTRIBUTE_NORMAL, ByVal 0&)
    If hClient = INVALID_HANDLE_VALUE Then
        WriteLine "#CreateFile failed, error " & CStr(Err.LastDllError)
    Else
        bytMsg = StrConv(Left$(Destination & vbTab _
                             & strUserName & " on " & strComputerName & vbTab _
                             & ReJoinAt(MsgStart), _
                               424), _
                         vbFromUnicode)
        If WriteFile(hClient, bytMsg(0), UBound(bytMsg) + 1, cbWritten, ByVal 0&) = 0 Then
            If Err.LastDllError = 53 Then
                WriteLine "#No network path to requested destination"
            Else
                WriteLine "#WriteFile failed, error " & CStr(Err.LastDllError)
            End If
        End If
        CloseHandle hClient
    End If
End Sub

Private Sub Usage()
    If Not blnInteractive Then
        WriteLine "#Send messages to LANegram."
        WriteLine
        WriteLine App.EXEName & " [/? | [/C <channel>] <name> <message>]"
    Else
        WriteLine "#Commands: [/? | [/C <channel>] <name> <message>]"
    End If
    WriteLine
    WriteLine "  /?           this help."
    WriteLine "  /C <channel> optional ""channel"" (no spaces), default is 0."
    WriteLine "  <name>       user or computer in primary domain/workgroup,"
    WriteLine "               optionally wildcarded using ? and *.  For example"
    WriteLine "               using * means send to all computers and users."
    WriteLine "  <message>    message text."
    If blnInteractive Then
        WriteLine
        WriteLine "  Empty input exits."
    End If
End Sub

Private Sub WriteChars(ByVal Text As String)
    Dim bytText() As Byte
    Dim cbWritten As Long
    
    bytText = StrConv(Text, vbFromUnicode)
    WriteFile hStdOut, bytText(0), UBound(bytText) + 1, cbWritten, ByVal 0&
End Sub

Private Sub WriteLine(Optional ByVal Text As String = "")
    WriteChars Text & vbNewLine
End Sub

Private Sub Main()
    Dim strCommand As String
    
    hStdOut = GetStdHandle(STD_OUTPUT_HANDLE)
    GetNames
    
    strCommand = Trim$(Command$())
    If Len(strCommand) > 0 Then
        ParseAndExecute strCommand
    Else
        blnInteractive = True
        WriteLine "#Enter commands, empty input to exit:"
        hStdIn = GetStdHandle(STD_INPUT_HANDLE)
        Do
            WriteChars ">"
            strCommand = Trim$(ReadLine())
            If Len(strCommand) > 0 Then
                ParseAndExecute strCommand
            End If
        Loop While Len(strCommand) > 0
    End If
End Sub
